/*
 * Copyright (C) 2005-2006 Atmel Corporation
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */
//#include <common.h>

//#include <asm/sysreg.h>
//#include <asm/ptrace.h>

#include "com.h"
#include "ptrace.h"
#include "sysreg.h"


extern unsigned long _stack;
extern unsigned long _estack;


//DECLARE_GLOBAL_DATA_PTR;
/*
 * 1 0xA000_0000 Reset External input Undefined
2 Provided by OCD system OCD Stop CPU OCD system First non-completed instruction
3 EVBA+0x00 Unrecoverable exception Internal PC of offending instruction
4 EVBA+0x04 TLB multiple hit Internal signal PC of offending instruction
5 EVBA+0x08 Bus error data fetch Data bus First non-completed instruction
6 EVBA+0x0C Bus error instruction fetch Data bus First non-completed instruction
7 EVBA+0x10 NMI External input First non-completed instruction
8 Autovectored Interrupt 3 request External input First non-completed instruction
9 Autovectored Interrupt 2 request External input First non-completed instruction
10 Autovectored Interrupt 1 request External input First non-completed instruction
11 Autovectored Interrupt 0 request External input First non-completed instruction
12 EVBA+0x14 Instruction Address ITLB PC of offending instruction
13 EVBA+0x50 ITLB Miss ITLB PC of offending instruction
14 EVBA+0x18 ITLB Protection ITLB PC of offending instruction
15 EVBA+0x1C Breakpoint OCD system First non-completed instruction
16 EVBA+0x20 Illegal Opcode Instruction PC of offending instruction
17 EVBA+0x24 Unimplemented instruction Instruction PC of offending instruction
18 EVBA+0x28 Privilege violation Instruction PC of offending instruction
19 EVBA+0x2C Floating-point FP Hardware PC of offending instruction
20 EVBA+0x30 Coprocessor absent Instruction PC of offending instruction
21 EVBA+0x100 Supervisor call Instruction PC(Supervisor Call) +2
22 EVBA+0x34 Data Address (Read) DTLB PC of offending instruction
23 EVBA+0x38 Data Address (Write) DTLB PC of offending instruction
24 EVBA+0x60 DTLB Miss (Read) DTLB PC of offending instruction
25 EVBA+0x70 DTLB Miss (Write) DTLB PC of offending instruction
26 EVBA+0x3C DTLB Protection (Read) DTLB PC of offending instruction
27 EVBA+0x40 DTLB Protection (Write) DTLB PC of offending instruction
28 EVBA+0x44 DTLB Modified DTLB PC of offending instruction
 *
 */
const char		*exceptions[] =
	{
		"Unrecoverable exception",
		"TLB multiple hit",
		"TLB multiple hit",
		"Bus error instruction fetch",
		"NMI External input",
		"Instruction Address ITLB",
		"ITLB Protection",
		"Breakpoint OCD",
		"Illegal Opcode Instruction",
		"Unimplemented instruction",
		"Privilege violation",
		"Floating-point",
		"Coprocessor absent Instruction",
		"Data Address (Read)",
		"Data Address (Write)",
		"DTLB Protection (Read)",
		"DTLB Protection (Write)",
		"DTLB Modified",
		"",
		"",
		"",
		"ITLB Miss",
		0
	};

static const char * const cpu_modes[8] = {
	"Application", "Supervisor", "Interrupt level 0", "Interrupt level 1",
	"Interrupt level 2", "Interrupt level 3", "Exception", "NMI"
};
/*
static void dump_mem(const char *str, unsigned long bottom, unsigned long top)
{
	unsigned long p;
	int i;

	Printf("%s(0x%08lx to 0x%08lx)\n", str, bottom, top);

	for (p = bottom & ~31; p < top; ) {
		Printf("%04lx: ", p & 0xffff);

		for (i = 0; i < 8; i++, p += 4) {
			unsigned int val;

			if (p < bottom || p >= top)
				Printf("         ");
			else {
				val = *(unsigned long *)p;
				Printf("%08x ", val);
			}
		}
		Printf("\n");
	}
}
*/
void do_unknown_exception(unsigned int ecr, struct pt_regs *regs)
{
	unsigned int mode;

	Printf("\n *** Unhandled exception %u (0x%x) at PC=0x%08lx: %s\n",	ecr, ecr << 2, regs->pc, exceptions[ecr]);

	switch (ecr) {
	case ECR_BUS_ERROR_WRITE:
	case ECR_BUS_ERROR_READ:
		Printf("Bus error at address 0x%08lx\n", sysreg_read(BEAR));
		break;
	case ECR_TLB_MULTIPLE:
	case ECR_ADDR_ALIGN_X:
	case ECR_PROTECTION_X:
	case ECR_ADDR_ALIGN_R:
	case ECR_ADDR_ALIGN_W:
	case ECR_PROTECTION_R:
	case ECR_PROTECTION_W:
	case ECR_DTLB_MODIFIED:
	case ECR_TLB_MISS_X:
	case ECR_TLB_MISS_R:
	case ECR_TLB_MISS_W:
		Printf("MMU exception at address 0x%08lx\n", sysreg_read(TLBEAR));
		Printf("MMU TLBEHI 0x%08lx\n", sysreg_read(TLBEHI));
		Printf("MMU TLBELO 0x%08lx\n", sysreg_read(TLBELO));
		Printf("MMU PTBR 0x%08lx\n", sysreg_read(PTBR));
		Printf("MMU MMUCR 0x%08lx\n", sysreg_read(MMUCR));
		break;
	}

	Printf("   pc: %08lx    lr: %08lx    sp: %08lx   r12: %08lx\n",
	       regs->pc, regs->lr, regs->sp, regs->r12);
	Printf("  r11: %08lx   r10: %08lx    r9: %08lx    r8: %08lx\n",
	       regs->r11, regs->r10, regs->r9, regs->r8);
	Printf("   r7: %08lx    r6: %08lx    r5: %08lx    r4: %08lx\n",
	       regs->r7, regs->r6, regs->r5, regs->r4);
	Printf("   r3: %08lx    r2: %08lx    r1: %08lx    r0: %08lx\n",
	       regs->r3, regs->r2, regs->r1, regs->r0);
	Printf("Flags: %c%c%c%c%c\n",
	       regs->sr & SR_Q ? 'Q' : 'q',
	       regs->sr & SR_V ? 'V' : 'v',
	       regs->sr & SR_N ? 'N' : 'n',
	       regs->sr & SR_Z ? 'Z' : 'z',
	       regs->sr & SR_C ? 'C' : 'c');
	Printf("Mode bits: %c%c%c%c%c%c%c%c%c\n",
	       regs->sr & SR_H ? 'H' : 'h',
	       regs->sr & SR_R ? 'R' : 'r',
	       regs->sr & SR_J ? 'J' : 'j',
	       regs->sr & SR_EM ? 'E' : 'e',
	       regs->sr & SR_I3M ? '3' : '.',
	       regs->sr & SR_I2M ? '2' : '.',
	       regs->sr & SR_I1M ? '1' : '.',
	       regs->sr & SR_I0M ? '0' : '.',
	       regs->sr & SR_GM ? 'G' : 'g');
	mode = (regs->sr >> SYSREG_M0_OFFSET) & 7;
	Printf("CPU Mode: %s\n", cpu_modes[mode]);

	/* Avoid exception loops */
	if (regs->sp < _stack || regs->sp >= _estack)
		Printf("\nStack pointer seems bogus, won't do stack dump\n");
//	else
//		dump_mem("\nStack: ", regs->sp, _estack);

	Prints("Unhandled exception\n");
}
